#define _MULTI_THREADED

#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>
#include <pthread.h>
#include <signal.h>

#include "watek.h"
#include "gui.h"

const int OK = 0;
const int ZLE_HASLO = 1;
const int SERWER_PELNY = 2;
const int BLAD = 3;
const int ANULOWANIE = 4;
const int WYSYLANIE = 5;
const int UAKTUALNIENIE = 6;

extern "C" void *threadStart(void *);
extern "C" void sigAction(int);

WatekKlienta* WatekKlienta::watek;

WatekKlienta::WatekKlienta(GUI&g, int argc, char** argv):gui(g) {
	port = 1023;
	
	id = pthread_self();
	watek = this;
	  
	if (argc<3) {
    		printf("Parametry: host port haslo\n");
    		::exit(EXIT_FAILURE);
  	}

	sscanf(argv[2], "%d", &port);
 	if (port<1024 || port>65535) {
    		fprintf(stderr, "Nr portu musi byc z przedzialu 1024-65535\n");
    		::exit(EXIT_FAILURE);
  	}

  	if (strlen(argv[3])>16) {
    		fprintf(stderr, "Maksymalna dlugosc hasla to 16 znakow\n");
    		::exit(EXIT_FAILURE);
  	}

  	sd = socket(PF_INET, SOCK_STREAM, 0);        
    
	struct hostent* host = gethostbyname(argv[1]);
	if (host==0) {
		herror("gethostbyname");
        	::exit(EXIT_FAILURE);
	}
	
	memset(&adr, 0, sizeof(struct sockaddr_in));	
    	adr.sin_family = AF_INET;
    	adr.sin_port = htons(port);
	adr.sin_addr.s_addr = inet_addr(inet_ntoa(*((struct in_addr *)host->h_addr)));
	
    	if (connect(sd, (struct sockaddr *)&adr, sizeof(struct sockaddr))) {
    		perror("connect");
      		close(sd);
      		::exit(EXIT_FAILURE);
	}

	printf("Polaczeno\n");
	
    	write (sd, argv[3], strlen(argv[3])+1);

    	read(sd, &msg, sizeof(int));
    
    	if (msg == ZLE_HASLO) {
      		printf("Podane haslo jest nieprawidlowe\n");
      		close(sd);
      		::exit(EXIT_FAILURE);
    	} else if (msg == SERWER_PELNY) {
      		printf("Serwer pelny\n");
      		close(sd);
      		::exit(EXIT_FAILURE);
	}

	printf("Haslo zaakceptowane\n");
	
    	recvTxt();
		
	gui.nowa();
	
	signal(SIGUSR1, sigAction);
	
	pthread_t threadid;
	pthread_create(&threadid, 0, threadStart, this);
}

void WatekKlienta::recvTxt() {
	unsigned long len;
	
	printf("Transfer tekstu:\n");
	
    	read(sd, &len, sizeof(unsigned long));
	
	printf("  naglowek\n");
	
    	unsigned long pos = 0;
	txt = (char*)malloc(len+1);
	txt[len]='\0';
	
	printf("  tekst []");
	
    	while(pos!=len) {
      		memset(buf, '\0', 1025);
      		pos += read(sd, &(txt[pos]), len-pos>=1024 ? 1024 : len-pos);
		printf("\b.]");
    	}
	printf("\n");
	
	int cnt;
	tab[0] = tab[1] = 0;
	len=0;

	read(sd, &cnt, sizeof(int));
	fflush(stdout);
	
	if (pthread_self()==id) {
		gui.setTxt(txt);
	
		for(; cnt; cnt--) {
			read(sd, tab, 2*sizeof(unsigned long));
			gui.dodajRegion(tab[0], tab[1]);
		}
		gui.finish();
	} else {
		todo = 1;
		pthread_kill(id, SIGUSR1);
		while (todo);
		
		printf("  fragmentow w edycji: %d (", cnt);
		for(; cnt; cnt--) {
			read(sd, tab, 2*sizeof(unsigned long));
			printf(" %d %d", tab[0], tab[1]);
			todo = 2;
			pthread_kill(id, SIGUSR1);
			while (todo);
		}
		
		printf(" )\n");
		
		todo = 3;
		pthread_kill(id, SIGUSR1);
		while (todo);
	}
	
	fflush(stdout);
}

void WatekKlienta::akceptacja(unsigned long pocz, unsigned long kon) {
	write(sd, "akcp", sizeof(int));
	write(sd, &pocz, sizeof(unsigned long));
	write(sd, &kon, sizeof(unsigned long));
}

void WatekKlienta::anulowanie() {
	write(sd, &ANULOWANIE, sizeof(int));
}

void WatekKlienta::wyslij() {
	write(sd, &WYSYLANIE, sizeof(int));
	unsigned long len = gui.getBufTxt().length();
	write(sd, &len, sizeof(unsigned long));
	writeall(sd, gui.getBufTxt().local8Bit(), len);
}

int WatekKlienta::writeall(int s, const char*buf, unsigned long len) {
  int total = 0;
  int bytesleft = len;
  int n;

  while(total < len) {
    n = write(s, &buf[total], bytesleft);
    if (n == -1) { break; }
    total += n;
    bytesleft -= n;
  }

  len = total;

  return n==-1?-1:0;
} 

void WatekKlienta::run() {
	int msg;
	
	for (;;) {
	if (!read(sd, &msg, sizeof(int))) {
		todo = 5;
		pthread_kill(id, SIGUSR1);
		while(todo);
		close(sd);
		break;
	} else {
		if (msg == OK) {
			todo = 4;
			pthread_kill(id, SIGUSR1);
			while (todo);
			recvTxt();
		} else if (msg == UAKTUALNIENIE) {
			read(sd, &len, sizeof(long));
			todo = 7;
			pthread_kill(id, SIGUSR1);
			while (todo);
			recvTxt();
		} else {
			todo = 6;
			pthread_kill(id, SIGUSR1);
			while(todo);
		}
	}
	}
}

void WatekKlienta::zapisz(const char* co, QString gdzie) {
	int d = open(gdzie, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
	
	if (d==-1) {
		perror(0);
		return;
	}
	
	writeall(d, co, strlen(co));
	close(d);
}

void WatekKlienta::sygnal() {
	switch(todo) {
		case 1:
			gui.setTxt(txt);
			break;
		case 2:
			gui.dodajRegion(tab[0], tab[1]);
			break;
		case 3:
			gui.finish();
			break;
		case 4:
			if (gui.getStan() == OCZEKIWANIE) {
				gui.setStatus(QString("Serwer zaakceptowa")+QString(QChar(322))+QString(" nowy fragment"));
				gui.edycja();
			}
			break;
		case 5:
			gui.koniec();
			break;
		case 6:
			gui.brakAkcpt();
			break;
		case 7:
			gui.setZmianaDl(len);
			break;
	}
	todo = 0;
}

WatekKlienta* WatekKlienta::getWatek() {
	return watek;
}

void *threadStart(void *obiekt) {
  ((WatekKlienta*)obiekt)->run();
}

void sigAction(int) {
	WatekKlienta::getWatek()->sygnal();
}
